/* 
    added by Joerg Schulenburg 2006
    to switch between super computers (with slow disk and enough memory) 
    and small PCs (where we dont have enough memory) define USE_FILE
    to store and read the terms sequentially
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "Externals.h"
#include "FunctionPrototypes.h"
extern int myid;

#include "Constants.h"
#if EXPERIMENTAL & 2
#include <unistd.h>
extern int pipe_dat[2], pipe_cmd[2]; // pipes between mother and child=obtain
extern FILE *fdat0, *fdat1; // buffered streams
#ifdef USE_FILE
#error USE_FILE not allowed
#endif
#endif

//
// id: 0=gs_terms, 1=terms, 2=exp_terms 3=exc_terms
//
// because we have a small number (?) of different coefficients we
// can save memory by creating an index 
//   this is not for sure, so we only test it, please report to Joerg
//   if it will not work
// JS 2010-04-21
// MaxCoeffs=65535 should work mostly, but MaxCoeffs=255 is more compact (LC)
#define MaxCoeffs 65535               // not sure that we have a fixed maximum
double tableCoeffs[MaxCoeffs]={0.0}; // 0.0 marks the end!?
int    numCoeffs=0;

#ifdef USE_FILE
FILE *fp[4]={NULL,NULL,NULL,NULL};
#define  BUFSIZE (1024*1024l*8)
char buf[4][BUFSIZE];           // file cache
const char *name_ext[4]={"_gs",".terms",".exp_terms",".exc_terms"};
#else // !USE_FILE
// better have a growing list of memory blocks instead of realloc
//   realloc needs: Memory=oldmem+newmem and additional Memcopy, Jun08
#define ADDLEN (1024*1024l)     // num of terms per malloc block = 32MB
#define NUMBUFS     (1024l)     // maxMalloc = 32 * ADDLEN * NUMBUFS = 32GB
#ifndef TermCompression
// if TermCompression 1 energy and bra[x] may give less accuracy at 1e-7
#define TermCompression 0
#endif

// most compact version (bitshift + index) will be default in future
#if TermCompression                    // compress from 32B to 24B or 16B
# if TermCompression & 2
#  define DESCSIZE ((termUints)*sizeof(unsigned int)) // in bytes
#  if MaxCoeffs>255
#   define MPLYSIZE 2      //  multiplier size   (2 Byte) = 9..16
#  else
#   define MPLYSIZE 1      //  multiplier size   (1 Byte) = 5..8
#  endif
# else
#  define DESCSIZE ((termUints)*sizeof(unsigned int)) // in bytes
#  define MPLYSIZE (sizeof(float))      //  multiplier size   (4 Byte) = 16
# endif
#else
# define DESCSIZE (sizeof(int)*6)      // descriptors size (6*4 Byte)
# define MPLYSIZE (sizeof(double))     //  multiplier size   (8 Byte) = 32
#endif
#define TERMSIZE (DESCSIZE+MPLYSIZE)  //    one term size  (32/24/16 Byte)
int termUints=1+1+4; // number of uints needed to store all descriptor-bits
int maxConfBits, uintBits; // needed for best runtime compression
// t_buf: list of storage buffer pointers, only first elements initialized
char   *t_buf[4][NUMBUFS+1]={{0},{0},{0},{0}}; // first 4 NULL pointer
int     t_blk[4]={-2,-2,-2,-2}; // index(*char) to buffer block
size_t  t_idx[4]={~(size_t)0,~(size_t)0,~(size_t)0,~(size_t)0}; 
                  // at work: pointer to next element [in block]
                  // initially: >= ADDLEN(20bit)*TERMSIZE() or t_blk<0
#endif // !USE_FILE

/* id: 0=gs_terms 1=terms 2=exp_terms 3=exc_terms */
#ifdef USE_FILE
// static // not called from outside!
int term_open(int id, char *opt){
  char name[256];
  sprintf(name,"%s%s%_%04d", Control.name, name_ext[id], myid);
  if (myid==0) printf(" fopen(%s) - slow USE_FILE mode!\n",name);
  fp[id]=fopen(name,opt);
  if (!fp[id]) {
    printf("Cannot Open %s for %s...\nExiting!!\n\n", name, opt);
    ERR(1);
  }
  setvbuf(fp[id],buf[id],_IOFBF,BUFSIZE);
  return 0;
}
#endif

// JS 2010-04 get bits to store
//   max(NumberConfigurations, NumberExcitedConfigurations) <= size(int)
void term_rewind(int id) {  // called by obtain* + solve.c as initialization
  static int output=1; // JS1002 
  if (output) { output=0;
    int maxConf;
    unsigned int maxUint;
    maxConf=((NumberExcitedConfigurations>NumberConfigurations)?
              NumberExcitedConfigurations:NumberConfigurations );
    for(maxConfBits=0;maxConf;maxConf>>=1,maxConfBits++);
    for(maxUint=~0,uintBits=0;maxUint;maxUint>>=1,uintBits++);
// int termUints=1+1+4; 
    // number of uints needed to store all descriptor-bits
    termUints=(maxConfBits*5+7+uintBits-1)/uintBits;
    if (!myid)
      printf("\nTERMSIZE/Byte= %d  maxConfBits= %d\n\n",
             (int) TERMSIZE, maxConfBits);
    // see below static array definition d[6] as termUints max JS-2010-04-09
    if (termUints>6) { ERR("termUints>6"); exit(1); }
  }
  { static int old_nCoeffs=0;
    if (numCoeffs!=old_nCoeffs) {
      old_nCoeffs=numCoeffs;
      if (!myid)
        printf("DBG: term_rewind: numCoeffs= %d\n", numCoeffs);
    }
  }
//  printf("DBG: rewind %d\n", id);
#if defined(USE_FILE)
  if (!fp[id]) { term_open(id,"r"); }
  rewind(fp[id]);
#else
#if EXPERIMENTAL & 2
  { // except GS
    char buf=id;
    // ToDo JS0911: generates lot of syscalls which is slow, use bufferring!
    write(pipe_cmd[1],&buf,1); // trigger obtain in child
    return;
  }
#endif
  if (t_blk[id]==-2) t_buf[id][0]=0; // initial mark as free JS091120
  // cause initialization to t_buf[id][0] on first access
  t_blk[id]=-1;
  t_idx[id]=TERMSIZE*ADDLEN;
#endif
}

/* set error if storing was not possible,
   skip further storing, break getting  */
int error=0;

/*
 *  gs_terms:  id=0
 *  terms:     id=1
 *  exp_terms: id=2 (ToDo) Jun08
 *  ToDo: rename descriptors to cfgidxs or term_descriptor
 */
int term_store(int id, int *descriptors, double multiplier) {
#ifdef USE_FILE
  int nio;
  if (error) return 0; // skip on error
  if (!fp[id]) { term_open(id,"w+"); }
  nio =fwrite(descriptors,sizeof(   int),6,fp[id]);
  nio+=fwrite(&multiplier,sizeof(double),1,fp[id]);
  if (nio!=7) {
    error=1;
    printf("%3d: term_store fwrite_error=%d L%d\n",
                 myid, ferror(fp[id]), __LINE__);
    printf(" Terms can not be stored, you will get no results!\n"
           " Program will stop after all terms are counted.\n");
  }
#else // !USE_FILE
#if EXPERIMENTAL & 2  // pipe+threads
  { // except GS
    int nio;
#if 0
    nio = write(pipe_dat[1],descriptors,6*sizeof(int));
    nio+= write(pipe_dat[1],&multiplier,  sizeof(double));
    if (nio!=6*sizeof(int)+sizeof(double)) {
#else
    nio =fwrite(descriptors,sizeof(   int),6,fdat1);
    nio+=fwrite(&multiplier,sizeof(double),1,fdat1);
    if (descriptors[0]<0) fflush(fdat1); // must be send through pipe
    if (nio!=7) {
#endif
      error=1;
      printf("%3d: term_store\n", myid); perror("write_pipe");
    }
    return 0;
  }
#endif // EXPERIMENTAL & 2 (pipe+threads)
  if (error) return 0; // skip on error (dont store, but count)
  if   (t_idx[id] >= TERMSIZE*ADDLEN) {
    if (t_blk[id] == NUMBUFS-1) {
         printf("WARNING: not enough memory (NUMBUFS exceeded)\n"); }
    t_blk[id]++;        // extend the memory buffer
    t_idx[id]=0;
    t_buf[id][ t_blk[id]+1 ]=0;  // mark as not allocated (for term_free)
    // realloc has the problem of fragmentation and copying (max RAM/2)
    t_buf[id][ t_blk[id] ]=(char *)malloc(TERMSIZE*ADDLEN);
//  printf("DBG: malloc block id= %d blk= %d\n", id, t_blk[id]);
    if (!t_buf[id][ t_blk[id] ]) {
      printf("%4d: malloc failed, bkl= %10.2e [blklen=%.0f*%.2e]\n",
             myid, 1.*t_blk[id], 1.*TERMSIZE, 1.*ADDLEN);
      printf(" see " __FILE__ " L%d for more info!\n", __LINE__);
      // Further terms can not be stored, you will get no results!
      // Program will stop after all terms are counted.
      // If you have fast storage, try USE_FILE which is slower
      // you can also use the PIPELINED version or Term compression
#if 1 // set to 1 for debugging
      error=system("free;ps ux;ulimit -a");
      // system("sleep 60");  // order the output?
#endif
      error=2;
      return -1;
    }
  }
#if TermCompression
  // int maxConfBits, uintBits; // needed for best runtime compression
# if TermCompression & 2
  int m=0;
#else
  float m=multiplier;
#endif
  // static array definition d[6] as termUints max JS-2010-04-09
  // use 6+1 for simpler code
  unsigned int d[6+1]={0,0,0,0,0,0,0}; // zero for every run!
  int usedBits=0; // 0..uintBits
  int usedUint=0; // 0..6         sum_usedBits = usedUint*uintBits+usedBits
  int data, dataBits, sBits; // data + data bits to split to d[i+0] and d[i+1]
  int j;
//#define MaxCoeffs 255               // not sure that we have a fixed maximum
//double tableCoeffs[MaxCoeffs]={0.0}; // 0.0 marks the end!?
//int    numCoeffs=0;
//  not used at the moment
  if (numCoeffs<MaxCoeffs) { // should be always the case
    int i;
    for (i=0;i<numCoeffs;i++) if (tableCoeffs[i]==multiplier) break;
    if (i==numCoeffs) {  // not found -> add new entry
      tableCoeffs[i]=multiplier; numCoeffs++;
      { /* logarithmic debugging output */
        static int cBits=3; /* start from binary=1000 = 2^3 */
        if (numCoeffs==1<<cBits) {
          cBits++;
          if (!myid)
            printf("DBG: term_store: numCoeffsBits= %d\n", cBits);
        }
      }
      if (numCoeffs==MaxCoeffs && !myid)
        printf("\nERROR: numCoeffs>%d, please report to Joerg!!!\n\n",
               MaxCoeffs);
#if TermCompression & 2
      if (numCoeffs==MaxCoeffs) ERR("MaxCoeffs to small!?");
#endif
    }
# if TermCompression & 2
    m=i;
#endif
  }
  // put descriptor[0..6] bitwise to uint words (compression)
  d[0]=0;
  for (j=0;j<6;j++) {
    // j=0: cfg0, j=1: term(7bit), j=2..6: cfg1..4
    data= descriptors[j]; dataBits= ((j==1)?7:maxConfBits);
    while (dataBits) {
      // compute bits fitting to the word d[i]
      sBits = ((dataBits<=uintBits-usedBits)?dataBits:uintBits-usedBits);
      d[usedUint] |= data<<usedBits;      // store sbits from data to d[]
      dataBits -= sBits;  data >>= sBits; // remove stored sbits from data
      usedBits += sBits;
      if (usedBits==uintBits) { usedUint++; usedBits=0; d[usedUint]=0; }
    }
  }
  memcpy(t_buf[id][t_blk[id]]+t_idx[id],  d, DESCSIZE); t_idx[id]+=DESCSIZE;
  memcpy(t_buf[id][t_blk[id]]+t_idx[id], &m, MPLYSIZE); t_idx[id]+=MPLYSIZE;
#else // ~TermCompression
  memcpy(t_buf[id][t_blk[id]]+t_idx[id], descriptors, DESCSIZE); t_idx[id]+=DESCSIZE;
  memcpy(t_buf[id][t_blk[id]]+t_idx[id], &multiplier, MPLYSIZE); t_idx[id]+=MPLYSIZE;
#endif // TermCompression
#endif // USE_FILE
{
  static int nout=0;
  if (nout!=100 && Control.output>7 && !myid)
    printf("term_store %4d - %4d %2d . %4d %4d %4d %4d %7.3f\n", nout++,
     descriptors[0], descriptors[1],
     descriptors[2], descriptors[3], descriptors[4], descriptors[5],
     multiplier);
}
  return 0;
}

// this function may be usefull if memory is critical (max LSUB for LC)
int term_free(int id) {
#ifdef USE_FILE 
  // delete file
#else
#if EXPERIMENTAL & 2 
  // nothing too free
#endif
  int i;
  for (i=0;i<NUMBUFS+1 && t_buf[id][i];i++) free (t_buf[id][i]);
  t_buf[id][0]=0; // mark as free
#endif
  return 0;
}

// ToDo: JS0902  MPI_File_iread (async read)
// MPI_File_iwrite(fh, buf, count, datatype, &request);
// MPI_Wait(&request, &status);
// is equivalent to
// MPI_File_write(fh, buf, count, datatype, &status);
// and
// MPI_File_iread(fh, buf, count, datatype, &request);
// MPI_Wait(&request, &status);
// is equivalent to
// MPI_File_read(fh, buf, count, datatype, &status);
// replace 2 fread calls by one fread(struct) call? 

           /* id: 0=gs_terms, 1=terms */
// returning 0 on OK, else exit
int term_get(int id, int *descriptors, double *multiplier) {
  if (error) {
     printf("Error term_get: Not all terms could be stored!\n"
            "Error term_get: Take the counters to estimate what you need\n");
     ERR(1);
     // ToDo: repeat slow calculation of a single element here?
  }
#ifdef USE_FILE   // bad for MPI Cluster
  int nio;
  if (!fp[id]) { term_open(id,"r"); }
  nio =fread(descriptors,sizeof(   int),6,fp[id]);
  nio+=fread(multiplier ,sizeof(double),1,fp[id]);
  if (nio!=7) {
    printf("%3d: term_get fread_error=%d L%d\n",
                 myid, ferror(fp[id]), __LINE__); ERR(1); }
#else
#if EXPERIMENTAL & 2
  { // except GS
    int nio;
#if 0
    nio = read(pipe_dat[0],descriptors,6*sizeof(int));
    nio+= read(pipe_dat[0],multiplier ,  sizeof(double));
    if (nio!=6*sizeof(int)+sizeof(double)) {
#else
    nio =fread(descriptors,sizeof(   int),6,fdat0);
    nio+=fread(multiplier ,sizeof(double),1,fdat0);
    if (nio!=7) {
#endif
      printf(" %3d: read_pipe\n",myid); perror("read_pipe"); }
    return 0;
  }
#endif
  if (t_idx[id] >= TERMSIZE*ADDLEN) {
    t_blk[id]++; t_idx[id]=0;
//  printf("DBG: next block id= %d blk= %d\n", id, t_blk[id]);
    if (t_blk[id] >= NUMBUFS) {
      printf("ERROR %3d: access beyond t_buf\n",myid); ERR(1); }
    if (!t_buf[id][ t_blk[id] ]) {
      printf("ERROR myid= %3d term_get: no data, t_buf[id=%d][blk=%d]==0\n",
        myid, id, t_blk[id]);
      ERR("term_get");
    }
  }
//  if (t_idx[id]<1000) printf("DBG: term_get id= %d blk= %d idx= %d\n", id, t_blk[id], t_idx[id]);
#if TermCompression
# if TermCompression & 2
  int m=0;
#else
  float m;
#endif
  // static array definition d[6] as termUints max JS-2010-04-09
  unsigned int d[6]={0,0,0,0,0,0};
  int usedBits=0; // 0..uintBits
  int usedUint=0; // 0..6         sum_usedBits = usedUint*uintBits+usedBits
  int data, dataBits, sBits; // data + data bits to split to d[i+0] and d[i+1]
  int j, offsetBits;
  memcpy( d,t_buf[id][t_blk[id]]+t_idx[id], DESCSIZE); t_idx[id]+=DESCSIZE;
  memcpy(&m,t_buf[id][t_blk[id]]+t_idx[id], MPLYSIZE); t_idx[id]+=MPLYSIZE;
  // get descriptor[0..6] bitwise from uint words (decompression)
  for (j=0;j<6;j++) {
    // j=0: cfg0, j=1: term(7bit), j=2..6: cfg1..4
    data = 0; offsetBits = 0; dataBits= ((j==1)?7:maxConfBits);
    while (dataBits) {
      // compute bits fitting to the word d[i]
      sBits = ((dataBits<=uintBits-usedBits)?dataBits:uintBits-usedBits);
      data |= ((d[usedUint] >> usedBits) & ~((~0)<<sBits))<<offsetBits;
      dataBits -= sBits; offsetBits+= sBits;
      usedBits += sBits;
      if (usedBits==uintBits) { usedUint++; usedBits=0; }
    }
    descriptors[j] = data;
  }
# if TermCompression & 2
  *multiplier=tableCoeffs[m];
#else
  *multiplier=(double)m;
#endif
#else
  memcpy(descriptors,t_buf[id][t_blk[id]]+t_idx[id], DESCSIZE); t_idx[id]+=DESCSIZE;
  memcpy(multiplier ,t_buf[id][t_blk[id]]+t_idx[id], MPLYSIZE); t_idx[id]+=MPLYSIZE;
#endif
#endif
{
  static int nout=0;
  if (nout!=100 && Control.output>7 && !myid)
    printf("term_get %4d - %4d %2d . %4d %4d %4d %4d %7.3f\n", nout++,
     descriptors[0], descriptors[1],
     descriptors[2], descriptors[3], descriptors[4], descriptors[5],
     *multiplier);
}
  return 0;
}

// Test:
//  gcc -g -DTEST -Wall storage.c
//  valgrind -v --tool=memcheck ./a.out
#ifdef TEST
int myid=0;
int main() {
  int i;
  int x[6]={0,1,2,3,4,5};
  double y=1.2345;
  for (i=0;i<10;i++) term_store(0,x,y);
  for (i=0;i<11;i++) term_get(0,x,&y);
}
#endif

// (Joerg Jun08: term_get3 replaced by id=3!)
